#include <stdio.h>
#include <stdlib.h>

void generate_bitstream(int *bitstream, int n_bits) {
    for (int i = 0; i < n_bits; i++) {
        bitstream[i] = rand() % 2;
    }
}

// Function to generate Walsh matrix
void generate_walsh_matrix(int n, int N, int walsh_matrix[N][N]) {
    if (n == 1) {
        walsh_matrix[0][0] = 1;
        return;
    }

    int half_n = n / 2;
    generate_walsh_matrix(half_n, N, walsh_matrix);

    for (int i = 0; i < half_n; i++) {
        for (int j = 0; j < half_n; j++) {
            walsh_matrix[i + half_n][j] = walsh_matrix[i][j];  // Copy top-left quadrant to top-right quadrant
            walsh_matrix[i][j + half_n] = walsh_matrix[i][j];  // Copy top-left quadrant to bottom-left quadrant
            walsh_matrix[i + half_n][j + half_n] = -walsh_matrix[i][j]; // Copy negation of top-left quadrant to bottom-right quadrant
        }
    }
}

// Function to sort Walsh matrix
void sort_walsh_matrix(int n, int walsh_matrix[n][n]) {
    // calculate number of zero-crossings in each row
    int zero_crossings[n];
    for (int i = 0; i < n; i++) {
        zero_crossings[i] = 0;
        for (int j = 1; j < n; j++) {
            if (walsh_matrix[i][j] != walsh_matrix[i][j - 1]) {
                zero_crossings[i]++;
            }
        }
    }
    // sort rows based on number of zero-crossings
    for (int i = 0; i < n - 1; i++) {
        for (int j = i + 1; j < n; j++) {
            if (zero_crossings[i] > zero_crossings[j]) {
                // swap rows
                for (int k = 0; k < n; k++) {
                    int temp = walsh_matrix[i][k];
                    walsh_matrix[i][k] = walsh_matrix[j][k];
                    walsh_matrix[j][k] = temp;
                }
                // swap zero-crossings
                int temp = zero_crossings[i];
                zero_crossings[i] = zero_crossings[j];
                zero_crossings[j] = temp;
            }
        }
    }
}

// Function to modulate data using Walsh code
void modulate(int data[], int walsh_code[], int data_size, int code_size, int modulated_signal[]) {
    for (int i = 0; i < data_size; i++) {
        for (int j = 0; j < code_size; j++) {
            modulated_signal[i * code_size + j] = data[i] * walsh_code[j];
        }
    }
}

// Function to demodulate received signal using Walsh code
void demodulate(int received_signal[], int walsh_code[], int data_size, int code_size, int demodulated_data[]) {
    for (int i = 0; i < data_size; i++) {
        int sum = 0;
        for (int j = 0; j < code_size; j++) {
            sum += received_signal[i * code_size + j] * walsh_code[j];
        }
        demodulated_data[i] = sum > 0 ? 1 : -1;
    }
}

// Function to convert 0 to -1
void convert_zero_to_minus_one(int data[], int size) {
    for (int i = 0; i < size; i++) {
        if (data[i] == 0) {
            data[i] = -1;
        }
    }
}

// Function to convert -1 to 0
void convert_minus_one_to_zero(int data[], int size) {
    for (int i = 0; i < size; i++) {
        if (data[i] == -1) {
            data[i] = 0;
        }
    }
}

int main() {
    // Define parameters
    int data_size = 4;
    int code_size = 4; // Walsh code length (must be power of 2)
    int channels = 2; // Number of channels used
    // printf("Enter Walsh code length, Number of Channels and Data Size: ");
    scanf("%d %d %d", &code_size, &channels, &data_size);

    // Generate Walsh matrix
    int walsh_matrix[code_size][code_size];
    generate_walsh_matrix(code_size, code_size, walsh_matrix);

    sort_walsh_matrix(code_size, walsh_matrix);

    // Print Walsh matrix
    // printf("Walsh Matrix:\n");
    // for (int i = 0; i < code_size; i++) {
    //     for (int j = 0; j < code_size; j++) {
    //         printf("%d ", walsh_matrix[i][j]);
    //     }
    //     printf("\n");
    // }

    // Allocate memory for data and modulated signals using malloc
    int *data = (int *)malloc(data_size * sizeof(int) * channels);
    int *modulated_signal = (int *)malloc(data_size * code_size * sizeof(int));
    int *received_signal = (int *)malloc(data_size * code_size * sizeof(int));

    // set received_signal to 0 
    for (int i = 0; i < data_size * code_size; i++) {
        received_signal[i] = 0;
    }

    // Get data for each channel
    for (int i = 0; i < channels; i++) {
        // printf("Enter Data for Channel %d: ", i + 1);
        for (int j = 0; j < data_size; j++) {
            int bit;
            scanf("%1d", &bit);
            if (bit == 0 || bit == 1) {
                data[i * data_size + j] = bit;
            }
            else {
                generate_bitstream(data + i * data_size, data_size);
                break;
            }
        }        
    }

    // Convert 0 to -1
    convert_zero_to_minus_one(data, data_size * channels);

    // Modulate data and Combine modulated signals (assuming simple addition for this example)
    for (int i = 0; i < channels; i++) {
        modulate(data + i * data_size, walsh_matrix[i], data_size, code_size, modulated_signal);
        // Print modulated signal
        printf("Modulated Signal for Channel %d: ", i + 1);
        for (int j = 0; j < data_size * code_size; j++) {
            printf("%d ", modulated_signal[j]);
        }
        printf("\n");
        
        for (int j = 0; j < data_size * code_size; j++) {
            received_signal[j] += modulated_signal[j];
        }
    }

    // Print received signal
    printf("Received Signal: ");
    for (int i = 0; i < data_size * code_size; i++) {
        printf("%d ", received_signal[i]);
    }
    printf("\n");

    // Demodulate received signal
    int* demodulated_data = (int *)malloc(data_size * sizeof(int));
    for (int i = 0; i < channels; i++) {
        demodulate(received_signal, walsh_matrix[i], data_size, code_size, demodulated_data);
        // Convert -1 to 0
        convert_minus_one_to_zero(demodulated_data, data_size);
        // Print demodulated data
        printf("Demodulated Data for Channel %d: ", i + 1);
        for (int j = 0; j < data_size; j++) {
            printf("%d", demodulated_data[j]);
        }
        printf("\n");
    }   

    // Free allocated memory
    free(data);
    free(modulated_signal);
    free(received_signal);
    free(demodulated_data);

    return 0;
}
