#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include <time.h>

#define MU 4           // Bits per symbol (i.e., 16-QAM)
#define SNR_DB 25      // Signal to noise ratio in dB

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

typedef struct {
    int bits[MU];
    double complex symbol;
} QAM16;

// Function prototypes
void generate_random_bits(int *bits, int length);
void qam16_modulation(int *bits, double complex *symbols, int length);
void add_cyclic_prefix(double complex *in, double complex *out, int cp_len, int n);
void remove_cyclic_prefix(double complex *in, double complex *out, int cp_len, int n);
void qam16_demodulation(double complex *symbols, int *bits, int length);
double calculate_ber(int *transmitted_bits, int *received_bits, int length);
void idft(double complex *data, double complex *output, int n);
void dft(double complex *data, double complex *output, int n);
double complex find_closest_symbol(double complex point);

const QAM16 mapping_table[16] = {
    {{0,0,0,0}, -3-3*I},
    {{0,0,0,1}, -3-1*I},
    {{0,0,1,0}, -3+3*I},
    {{0,0,1,1}, -3+1*I},
    {{0,1,0,0}, -1-3*I},
    {{0,1,0,1}, -1-1*I},
    {{0,1,1,0}, -1+3*I},
    {{0,1,1,1}, -1+1*I},
    {{1,0,0,0},  3-3*I},
    {{1,0,0,1},  3-1*I},
    {{1,0,1,0},  3+3*I},
    {{1,0,1,1},  3+1*I},
    {{1,1,0,0},  1-3*I},
    {{1,1,0,1},  1-1*I},
    {{1,1,1,0},  1+3*I},
    {{1,1,1,1},  1+1*I}
};

int main() {
    srand(time(NULL));  // Seed for random number generation

    int K;
    printf("Enter the number of OFDM subcarriers (K): ");
    scanf("%d", &K);
    int CP_LEN = K / 4; // Length of cyclic prefix: 25% of the block

    int transmitted_bits_length = K * MU;
    int *transmitted_bits = (int *)malloc(transmitted_bits_length * sizeof(int));
    int *received_bits = (int *)malloc(transmitted_bits_length * sizeof(int));
    double complex *symbols = (double complex *)malloc((K / MU) * sizeof(double complex));
    double complex *ofdm_time = (double complex *)malloc((K + CP_LEN) * sizeof(double complex));
    double complex *ofdm_tx = (double complex *)malloc((K + CP_LEN) * sizeof(double complex));
    double complex *ofdm_rx = (double complex *)malloc((K + CP_LEN) * sizeof(double complex));
    double complex *ofdm_demod = (double complex *)malloc(K * sizeof(double complex));
    double ber;

    // Get Transmitted bits from user
    printf("Transmitted bits: ");
    for (int i = 0; i < transmitted_bits_length; i++) {
        scanf("%1d", &transmitted_bits[i]);
    }

    // 16-QAM Modulation
    qam16_modulation(transmitted_bits, symbols, transmitted_bits_length);
    // printf("Transmitted symbols: ");
    // for (int i = 0; i < K / MU; i++) {
    //     printf("%lf%+lfi ", creal(symbols[i]), cimag(symbols[i]));
    // }
    // printf("\n");

    // IDFT
    idft(symbols, ofdm_time, K);
    // printf("OFDM time domain: ");
    // for (int i = 0; i < K; i++) {
    //     printf("%lf%+lfi ", creal(ofdm_time[i]), cimag(ofdm_time[i]));
    // }
    // printf("\n");

    // Add cyclic prefix
    add_cyclic_prefix(ofdm_time, ofdm_tx, CP_LEN, K);
    // printf("OFDM with cyclic prefix: ");
    // for (int i = 0; i < K + CP_LEN; i++) {
    //     printf("%lf%+lfi ", creal(ofdm_tx[i]), cimag(ofdm_tx[i]));
    // }
    // printf("\n");

    // Channel
    for (int i = 0; i < K + CP_LEN; i++) {
        double noise_power = 0.5 / pow(10, SNR_DB / 10);
        double noise = sqrt(noise_power / 2) * (rand() % 2 ? 1 : -1);
        ofdm_rx[i] = ofdm_tx[i] + noise;
    }

    // Remove cyclic prefix
    remove_cyclic_prefix(ofdm_rx, ofdm_time, CP_LEN, K);
    // printf("OFDM received: ");
    // for (int i = 0; i < K; i++) {
    //     printf("%lf%+lfi ", creal(ofdm_time[i]), cimag(ofdm_time[i]));
    // }
    // printf("\n");

    // DFT
    dft(ofdm_time, ofdm_demod, K);
    // printf("OFDM demodulated: ");
    // for (int i = 0; i < K; i++) {
    //     printf("%lf%+lfi ", creal(ofdm_demod[i]), cimag(ofdm_demod[i]));
    // }
    // printf("\n");

    // 16-QAM Demodulation
    qam16_demodulation(ofdm_demod, received_bits,  transmitted_bits_length);
    printf("Received bits: ");
    for (int i = 0; i < transmitted_bits_length; i++) {
        printf("%d", received_bits[i]);
    }
    printf("\n");

    // Calculate Bit Error Rate (BER)
    ber = calculate_ber(transmitted_bits, received_bits, transmitted_bits_length);

    printf("Bit Error Rate: %lf\n", ber);

    // Free dynamically allocated memory
    free(transmitted_bits);
    free(received_bits);
    free(symbols);
    free(ofdm_time);
    free(ofdm_tx);
    free(ofdm_rx);
    free(ofdm_demod);

    return 0;
}

// Generate random bits
void generate_random_bits(int *bits, int length) {
    for (int i = 0; i < length; i++) {
        bits[i] = rand() % 2;
    }
}

// 16-QAM Modulation
void qam16_modulation(int *bits, double complex *symbols, int length) {
    for (int i = 0; i < length / MU; i++) { //for every subcarrier
        int index = bits[4*i] * 8 + bits[4*i+1] * 4 + bits[4*i+2] * 2 + bits[4*i+3];
        symbols[i] = mapping_table[index].symbol;
    }
}

// Add cyclic prefix
void add_cyclic_prefix(double complex *in, double complex *out, int cp_len, int n) {
    for (int i = 0; i < cp_len; i++) {
        out[i] = in[n - cp_len + i];
    }
    for (int i = 0; i < n; i++) {
        out[cp_len + i] = in[i];
    }
}

// Remove cyclic prefix
void remove_cyclic_prefix(double complex *in, double complex *out, int cp_len, int n) {
    for (int i = 0; i < n; i++) {
        out[i] = in[cp_len + i];
    }
}

// 16-QAM Demodulation
void qam16_demodulation(double complex *symbols, int *bits, int length) {
    for (int i = 0; i < length / MU; i++) {
        double complex closest = find_closest_symbol(symbols[i]);
        for (int j = 0; j < 16; j++) {
            if (mapping_table[j].symbol == closest) {
                bits[4*i] = mapping_table[j].bits[0];
                bits[4*i+1] = mapping_table[j].bits[1];
                bits[4*i+2] = mapping_table[j].bits[2];
                bits[4*i+3] = mapping_table[j].bits[3];
                break;
            }
        }
    }
}

// Calculate Bit Error Rate
double calculate_ber(int *transmitted_bits, int *received_bits, int length) {
    int error_count = 0;
    for (int i = 0; i < length; i++) {
        if (transmitted_bits[i] != received_bits[i]) {
            error_count++;
        }
    }
    return (double)error_count / length;
}

// IDFT
void idft(double complex *data, double complex *output, int n) {
    for (int k = 0; k < n; k++) {
        output[k] = 0;
        for (int m = 0; m < n; m++) {
            output[k] += data[m] * cexp(2 * M_PI * I * k * m / n);
        }
        output[k] /= n;
    }
}

// DFT
void dft(double complex *data, double complex *output, int n) {
    for (int k = 0; k < n; k++) {
        output[k] = 0;
        for (int m = 0; m < n; m++) {
            output[k] += data[m] * cexp(-2 * M_PI * I * k * m / n);
        }
    }
}

// Find closest symbol in mapping table
double complex find_closest_symbol(double complex point) {
    double min_dist = cabs(point - mapping_table[0].symbol);
    double complex closest_symbol = mapping_table[0].symbol;

    for (int i = 1; i < 16; i++) {
        double dist = cabs(point - mapping_table[i].symbol);
        if (dist < min_dist) {
            min_dist = dist;
            closest_symbol = mapping_table[i].symbol;
        }
    }

    return closest_symbol;
}
